(*
 *  beware ircd, Internet Relay Chat server, buser.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit buser;

interface

uses blinklist,{winsock,wsocket,}bcmds,bstuff,bconsts,bipcheck;

type
  tusermodetable=record
    c:char;               {mode char}
    flag:integer;         {mode flag}
  end;

const
  usermodesupported='dioswkg';
  usermode_invisible= $1;
  usermode_oper=      $2;
  usermode_wallops=   $4;
  usermode_notices=   $8;
  usermode_service=  $10;
  usermode_deaf=     $20;
  usermode_debug=    $40;
  usermode_vhost=   $100;


  maxusermodetable=7;

  {
  this table is used in sending/receiving N messages
  }
  usermodetable:array[0..maxusermodetable] of tusermodetable=(
  (c:'i';flag:usermode_invisible),
  (c:'o';flag:usermode_oper),
  (c:'w';flag:usermode_wallops),
  (c:'s';flag:usermode_notices),
  (c:'k';flag:usermode_service),
  (c:'d';flag:usermode_deaf),
  (c:'g';flag:usermode_debug),
  (c:'x';flag:usermode_vhost)
  );

  userflag_myuser=         $400;
  userflag_isserver=       $800;
  userflag_isclient=      $1000;
  userflag_destroying=    $2000;

  servflag_ulined=           $1; {server is U:lined}
  servflag_joining=          $2; {server doesnt have netburst completed (is J10)}
  servflag_destroying=       $4; {server is marked for being destroyed, ignore incoming SQUIT for this servre}
  servflag_nosquit=          $8; {don't send a SQUIT when destroying this server}
  servflag_destroying_root= $10;
  {set for the first server in markdestroyingserver. prevent the same server from being destroyed twice (crash)
  when a farther server gets destroyed first, and then a nearer server}

  userflag_nameset=    $1000000; {has name in name hashtable}

type
  pointerarray=array[0..0] of pointer;
  ppointerarray=^pointerarray;
  tserver=class(tlinklist)
    us:tlinklist;           {the user-object for this server}
    p10num:integer;         {numeric, 0..4095}
    p10slots:ppointerarray;
    p10max:integer;         {should be 2^n-1}
    parentserver:tserver;   {one server closer}
    usercount:integer;      {number of clients on server}
    linktime:integer;
    lag:integer;            {the lag number in /MAP}
    flags:integer;
    serverlinknum:integer;  {entry in server-link array}
    protoversion:integer;   {P##, other than 10 is accepted for remote servers}
    destructor destroy; override;
  end;

  tuser=class(tthing)

    userid:string;    {userid (supplied by identd or user command, identd overriding user command)}

    idstr:string;     {p10 numeric string SS for server or SSCCC for client}
    p10slotnum:integer;{entry in the server's P10 users array}

    host:string;      {the host as it shows in userhost}
    uhost:string;     {upcase host}
(*    ip:string;        {ip address  (get "host is resolved" by doing host=ip)}*)
    binip:integer;    {dword ip}
    fullname:string;

    modeflag:integer; {any boolean user modes}

    server:tserver;   {The Tserver this user is on. if this user isserver, it's his own server (not the parent)}
    hops:integer;     {hopcount}

    chancount:integer;  {on n channels}
    channel:tlinklist;

    flags:integer;    {other, internal, boolean vars, not modes}
    ts:integer;       {nick last changed, used for nick collision}

    account:string;          {services login name}
    srvunit:pointer;
    bottag:array[0..maxserviceunit] of pointer; {4 bytes for each service unit}
    ipcheck:tipcheck;
    destructor destroy; override;
  end;

{
function allocuser: new() a user and add it to the global user list
addlocaluser(p:tuser): add an existing user to be direct connection
}

var
  globaluserlist:tlinklist;
  me:tuser;      {the local server}
  hostservuser:tuser; {the server linked to me}

  count:record
    globalclients:integer;
    localclients:integer;
    unknown:integer;
    oper:integer;
    invisible:integer;
    globalservers:integer;
    localservers:integer;
    connections:integer;
    channels:integer;

    highestconnections:integer;
    highestlocalclients:integer;
    highestglobalclients:integer;
  end;


function adduser:tuser;
{procedure removeuser(us:tuser);}

function isunreg(us:tuser):boolean;
function isclient(us:tuser):boolean;
function isserver(us:tuser):boolean;
function myconnect(us:tuser):boolean;
function isoper(us:tuser):boolean;
function isprivileged(us:tuser):boolean;

function nickuserhost(us:tuser):string;

{checks if a name is occupied; also does jupes}
function nameinuse(s:string):boolean;
procedure setnickname(us:tuser;const s:string);


{a string of all this user's modes. note that if there are any modes set, the string is prefixed with a space}
function usermodestr(us:tuser):string;

function findnick(const s:string):tuser;
function findname(const s:string):tuser;
function finduser(const s:string;numeric:boolean):tuser;
function findnumeric(s:string):tuser;
function convertidstr(const s:string):string;

procedure setusermode(us,cptr,sptr:tuser;parc:integer;parv:pparams);


implementation

uses sysutils,bserver,bparse,bchannel,bsearchtree,bconfig,bservunit,bservicemngr;

var
  usertree:thashtable;

procedure setnickname(us:tuser;const s:string);
begin
  if flag_isset(us.flags,userflag_nameset) then deltree(@usertree,ircupper(us.name));
  us.name := s;
  if us.name <> '' then begin
    addtree(@usertree,ircupper(us.name),us);
    setflag(us.flags,userflag_nameset)
  end else
  clearflag(us.flags,userflag_nameset);
end;

function nameinuse;
begin
  s := ircupper(s);
  result := findname(s) <> nil;
end;

function adduser:tuser;
begin
  result := tuser.create;
  result.name := '';
  linklistadd(globaluserlist,tlinklist(result));
end;

function isclient;
begin
  result := flag_isset(us.flags,userflag_isclient);
end;

function isserver;
begin
  result := flag_isset(us.flags,userflag_isserver);
end;

function isunreg;
begin
  result := not (isserver(us) or isclient(us));
end;

function myconnect;
begin
  result := flag_isset(us.flags,userflag_myuser);
end;

function isoper;
begin
  result := flag_isset(us.modeflag,usermode_oper);
end;

function isprivileged;
begin
  result := flag_isset(us.modeflag,usermode_oper) or flag_isset(us.flags,userflag_isserver);
end;

destructor tuser.destroy;
var
  p,p2:tlinklist;
  a:integer;
begin
  if flag_isset(self.flags,userflag_destroying) then exit;
  setflag(self.flags,userflag_destroying);

  if self = hostservuser then begin
    
    hostservuser := nil;
    sock.close;
  end;

  if isclient(self) then begin
    {call ondestroyuser of serviceunit}

    for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
    if @serviceunitarray[a].ondestroyuser <> nil then
    serviceunitarray[a].ondestroyuser(self);

    {send quit to common}
    {remove from channels}
    while self.channel <> nil do begin
      deluserfromchannel(self,tuserchan(self.channel).ch,tuserchan(self.channel));
    end;
    if flag_isset(self.modeflag,usermode_invisible) then dec(count.invisible);
    if isoper(self) then dec(count.oper);
    dec(count.globalclients);

    {remove from P10 list}
    self.server.p10slots[self.p10slotnum] := nil;
    dec(self.server.usercount);

    if flag_isset(self.flags,userflag_myuser) then begin
      {my service bot, remove from botarray, undo things which addbot did}
      botarray[self.p10slotnum] := nil;
    end;
    ipcheck_disconnect(self.ipcheck);
  end;(* else *)

  if isserver(self) then begin
    markdestroyingserver(self.server);
    p := globalserverlist;
    while p <> nil do begin
      p2 := p.next;
      if tserver(p).parentserver = self.server then if (tserver(p) <> self.server) then begin
        if not flag_isset(tserver(p).flags,servflag_destroying_root) then begin
          setflag(tserver(p).flags,servflag_nosquit);
          tuser(tserver(p).us).destroy;
        end;
      end;
      p := p2;
    end;
    self.server.destroy;
    {dec count.servers done in tserver.destroy}
  end;

  setnickname(self,''); {remove name}
  linklistdel(globaluserlist,self);
  prev := nil;
  next := nil;
  inherited destroy;

end;

destructor tserver.destroy;
begin
  destroyserver(self);
  prev := nil;
  next := nil;
  inherited destroy;
end;

function nickuserhost;
begin
  if isclient(us) then result := us.name+'!'+us.userid+'@'+us.host
  else result := us.name
end;

function usermodestr(us:tuser):string;
var
  a:integer;
begin
  if us.modeflag = 0 then begin
    result := '';
    exit;
  end;
  result := ' +';
  for a := 0 to maxusermodetable do begin
    if us.modeflag and usermodetable[a].flag <> 0 then result := result + usermodetable[a].c;
  end;
end;

function convertidstr(const s:string):string;
begin
  {$ifdef shortnumerics}
  result := s;
  {$else}
  case length(s) of
    2,5: result := s;
    3: result := 'A'+s[1]+'A'+s[2]+s[3];
    1,4: result := 'A'+s;
  end;
  {$endif}
end;

function findnumeric(s:string):tuser;
var
  p:pointer;
  b:integer;
begin
  s := convertidstr(s);
  result := nil;
  if length(s) = SSCCClen then begin
    {SSCCC}
    b := p10strtoint(copy(s,1,SSlen));
    if b < 0 then exit;

    p := p10server[b];
    if p = nil then exit;
    b := p10strtoint(copy(s,1+SSlen,CCClen)); {3,3}
    if b < 0 then exit;
    p := tserver(p).p10slots[b and tserver(p).p10max];
    if p = nil then exit;
    if tuser(p).idstr = s then result := p
  end else if (length(s) = SSlen) then begin
    {SS}
    b := p10strtoint(s);
    if b < 0 then exit;
    p := p10server[b];
    if p <> nil then result := tuser(tserver(p).us);
  end
end;

function findname(const s:string):tuser;
begin
  result := findtree(@usertree,ircupper(s));
end;

function findnick(const s:string):tuser;
begin
  result := findname(s);
  if result <> nil then if not isclient(result) then result := nil;
end;

function finduser(const s:string;numeric:boolean):tuser;
begin
  if numeric then result := findnumeric(s)
  else result := findname(s);
  {}
end;

procedure setusermode(us,cptr,sptr:tuser;parc:integer;parv:pparams);
var
  cc:integer;
  setclear:integer;
  s:string;

procedure setmodechar(k:char;flag:integer;setclear:integer);
begin
  if setclear = 2 then begin
    us.modeflag := us.modeflag or flag;
  end else begin
    us.modeflag := us.modeflag and not flag;
  end;
end;

begin
  s := parv[2];
  setclear := 2;
  if us = nil then exit;
  if (parv[2] = '') or (parc < 3) then exit;
  for cc := 1 to length(s) do begin
    case s[cc] of
      '-':setclear := 1;
      '+':setclear := 2;
      else begin

          case s[cc] of
            'd':setmodechar('d',usermode_deaf,setclear);
            'i':begin
              if setclear = 2 then begin
                if not flag_isset(us.modeflag,usermode_invisible) then inc(count.invisible);
              end else begin
                if flag_isset(us.modeflag,usermode_invisible) then dec(count.invisible);
              end;
              setmodechar('i',usermode_invisible,setclear);
            end;
            'w':setmodechar('w',usermode_wallops,setclear);
            's':setmodechar('s',usermode_notices,setclear);
            'g':setmodechar('g',usermode_debug,setclear);
            'x':setmodechar('x',usermode_vhost,setclear);
            'o':begin
              if setclear = 2 then begin
                if not flag_isset(us.modeflag,usermode_oper) then inc(count.oper);
              end else begin
                if flag_isset(us.modeflag,usermode_oper) then dec(count.oper);
              end;
              setmodechar('o',usermode_oper,setclear);
            end;
            'k':setmodechar('k',usermode_service,setclear);
          end;

      end
    end;
  end;
end;


begin
  globaluserlist := nil;
end.
